
BEGIN {
  unless ($ENV{AUTHOR_TESTING}) {
    print "1..0 # SKIP these tests are for testing by the author\n";
    exit
  }
}

use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.007002
use Test::Spelling 0.12;
use Pod::Wordlist;


add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib  ) );
__DATA__
Alders
Alders'
Eilam
Eilam's
MaxMind
MaxMind's
Oschwald
Oschwald's
Rolsky
Rolsky's
API
APIs
Hartmaier
IANA
IP
IPv
TJ
distro
distro's
netmask
subnet
subnets
Dave
autarch
Greg
oschwald
Olaf
oalders
Inc
Alexander
abraxxa
drolsky
goschwald
Mather
tjmather
William
Stevenson
wstevenson
lib
Net
Works
Address
Network
Role
Types
Util
